import ext_llm as xllm
import concurrent.futures
#read config yaml file
config : str = open("config.yaml").read()
#initialize extllm library


extllm = xllm.init(config)

llm_client = extllm.get_model("groq-llama")
llm_streaming_client = extllm.get_model("groq-llama-streaming")
custom_client = extllm.get_model("custom-provider")
#non blocking calls

print("first call")
future1 = llm_client.generate_text("Sei un assistente", "Recitami il primo articolo della costituzione italiana")
print("second call")
future2 = llm_streaming_client.generate_text("Sei un assistente", "Recitami il primo emendamento della costituzione americana")
# non blocking calls

print("third call on user implemented llm client class")
future3 = custom_client.generate_text("Sei un assistente", "Recitami il primo emendamento della costituzione americana")
print("-------------------waiting for third result------------------")
print(future3.result())

print("-------------------waiting for first result-------------------")
print(future1.result())
print("-------------------waiting for second result------------------")
stream = future2.result()
for event in stream:
    print(event)


import concurrent.futures
from concurrent.futures import Future

from ext_llm import Llm
from ext_llm.llm.llm import Stream, Response


class LlmConcurrencyProxy(Llm):

    def __init__(self, llm: Llm):
        super().__init__()
        self.llm = llm
        self.executor = concurrent.futures.ThreadPoolExecutor()

    def generate_text(self, system_prompt: str, prompt: str, max_tokens=None, temperature=None) -> Future[Stream | Response]:
        future = self.executor.submit(self.llm.generate_text, system_prompt, prompt, max_tokens, temperature)
        return future

    def get_config(self):
        return self.llm.get_config()


import yaml

class ExtLlmContext:
    config_map = {} # type: dict
    def __init__(self, content: str):
        #read yaml file with nested properties
        self.config_map = yaml.safe_load(content)

    def get_configs(self) -> dict:
        return self.config_map["config"]

    def get_config(self, key:str):
        return self.config_map["config"][key]


from .llm.llm import Llm
from .context import ExtLlmContext
from ext_llm.extllm import ExtLlm

# This function initializes the ExtLlm object
# It takes in a string containing the yaml configuration
# It returns an ExtLlm object
def init(configs: str) -> ExtLlm:
    ext_llm_context = ExtLlmContext(configs)
    return ExtLlm(ext_llm_context)


from ext_llm import Llm


class MyLlm(Llm):
    def __init__(self, config: dict):
        super().__init__()
        self.config = config

    def generate_text(self, system_prompt: str, prompt: str, max_tokens=None, temperature=None):
        return "Hello, world!"


import os

from ext_llm import Llm
import groq
from ext_llm.llm.llm import Stream, Response

class GroqLlm(Llm):

    def __init__(self, config: dict):
        super().__init__()
        self.config = config
        groq_api_key_env_var_name = config['groq_api_key_variable_name']
        groq_api_key = os.getenv(groq_api_key_env_var_name)
        if groq_api_key is None:
            raise ValueError(f"Environment variable {groq_api_key_env_var_name} not set")
        self.__groq_client = groq.Client(api_key=groq_api_key)

    def __invoke_model(self, system_prompt, prompt, max_tokens: int, temperature: float, top_p=0.9) -> Response | Stream:
        is_stream = self.config['invocation_method'] == 'converse_stream'
        chat_completion = self.__groq_client.chat.completions.create(
            messages = [
                {
                    "role": "system",
                    "content": system_prompt
                },
                {
                    "role": "user",
                    "content": prompt
                }
            ],
            model = self.config['model_id'],
            max_completion_tokens = max_tokens,
            temperature = temperature,
            top_p = top_p,
            stop=None,
            stream=is_stream
        )
        if is_stream:
            return GroqStream(chat_completion)
        else:
            return Response(chat_completion.choices[0].message.content, chat_completion)

    def generate_text(self, system_prompt : str, prompt : str, max_tokens=None, temperature=None) -> Stream | Response:
        if max_tokens is None:
            max_tokens = self.config['max_tokens']

        if temperature is None:
            temperature = self.config['temperature']

        if self.config['invocation_method'] != 'converse' and self.config['invocation_method'] != 'converse_stream':
            raise ValueError("Invalid invocation_method in config")
        return self.__invoke_model(system_prompt, prompt, max_tokens, temperature)

    def get_config(self):
        return self.config

class GroqStream(Stream):

    def __init__(self, stream):
        super().__init__()
        self.stream = stream

    def __iter__(self):
        for event in self.stream:
            yield Response(event.choices[0].delta, None)


from concurrent.futures import Future

from ext_llm import Llm
from ext_llm.llm.llm import Response, Stream


class LlmLoggingProxy(Llm):
    def __init__(self, llm: Llm):
        super().__init__()
        self.llm = llm

    def generate_text(self, system_prompt: str, prompt: str, max_tokens=None, temperature=None):
        #print(f"generate_text called with system_prompt: {system_prompt}, prompt: {prompt}, max_tokens: {max_tokens}, temperature: {temperature}\n")
        #do logging
        result = self.llm.generate_text(system_prompt, prompt, max_tokens, temperature)
        #do logging
        #print(f"generate_text returned: {result}\n")
        return result

    def get_config(self):
        return self.llm.get_config()


import logging
import os
from concurrent.futures import Future

import boto3

from ext_llm import Llm
from ext_llm.llm.llm import Response, Stream


def make_system_prompt(system_prompt):
    return [
            {
                "text": system_prompt
            }
        ]

def make_prompt(prompt):
    return [
                {
                    "role": "user",
                    "content": [
                        {
                            "text": prompt
                        }
                    ]
                }
            ]

def make_inference_config(max_tokens, temperature, top_p):
    return {
            "maxTokens": max_tokens,
            "temperature": temperature,
            "topP": top_p
        }

class AwsLlm(Llm):

    logger = logging.getLogger('AwsLlm')

    def __init__(self, config: dict):
        super().__init__()
        self.config = config
        access_key_id_env_var_name = config['aws_access_key_id_variable_name']
        secret_access_key_env_var_name = config['aws_secret_access_key_variable_name']
        access_key_id = os.getenv(access_key_id_env_var_name)
        secret_access_key = os.getenv(secret_access_key_env_var_name)
        if access_key_id is None:
            raise ValueError(f"Environment variable {access_key_id_env_var_name} not set")
        if secret_access_key is None:
            raise ValueError(f"Environment variable {secret_access_key_env_var_name} not set")

        region = config['aws_region']
        self.__session = boto3.Session(
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            region_name=region
        )
        self.__bedrock_client = self.__session.client(service_name='bedrock-runtime')
        self.__model_id = config['model_id']

    def __invoke_model(self, system_prompt, prompt, max_tokens: int, temperature: float, top_p=0.9) -> Stream | Response:
        system_prompts = make_system_prompt(system_prompt)
        messages = make_prompt(prompt)
        inference_config = make_inference_config(max_tokens, temperature, top_p)
        response = self.__bedrock_client.converse(modelId=self.__model_id, messages=messages, system=system_prompts, inferenceConfig=inference_config)
        return response['output']['message']['content'][0]['text']

    def __invoke_model_stream(self, system_prompt, prompt, max_tokens: int, temperature: float, top_p=0.9):
        system_prompts = make_system_prompt(system_prompt)
        messages = make_prompt(prompt)
        inference_config = make_inference_config(max_tokens, temperature, top_p)
        stream = self.__bedrock_client.converse_stream(modelId=self.__model_id, messages=messages, system=system_prompts, inferenceConfig=inference_config)
        return stream['stream']

    def generate_text(self, system_prompt: str, prompt: str, max_tokens=None, temperature=None) -> Stream | Response:
        if max_tokens is not None:
            max_tokens = self.config['max_tokens']

        if temperature is not None:
            temperature = self.config['temperature']

        if self.config['invocation_method'] == 'converse':
            return self.__invoke_model(system_prompt, prompt, max_tokens, temperature)
        elif self.config['invocation_method'] == 'converse_stream':
            return self.__invoke_model_stream(system_prompt, prompt, max_tokens, temperature)
        else:
            raise ValueError("Invalid invocation method")

    def get_config(self):
        return self.config

class AwsStream(Stream):

    def __init__(self, stream):
        super().__init__()
        self.stream = stream

    def __iter__(self):
        for event in self.stream:
            yield Response(event['delta']['content'][0]['text'], None)


from ext_llm import ExtLlmContext, Llm
from ext_llm.llm.concurrency_proxy import LlmConcurrencyProxy
from ext_llm.llm.logging_proxy import LlmLoggingProxy

class ExtLlm:

    provider_dict = {
        "groq" : ["ext_llm.llm.groq_llm", "GroqLlm"],
        "aws" : ["ext_llm.llm.aws_llm", "AwsLlm"]
    }

    def __init__(self, context: ExtLlmContext):
        self.context = context

    def list_available_models(self):
        return self.context.get_configs()["presets"]

    def get_model(self, preset_name: str, module_name="ext_llm.llm") -> Llm:
        class_name = None
        module_name = None
        if preset_name not in self.context.get_configs()["presets"]:
            raise Exception("Preset not found")

        if "provider" in self.context.get_configs()["presets"][preset_name]:
            class_name = self.provider_dict[self.context.get_configs()["presets"][preset_name]["provider"]][1]
            module_name = self.provider_dict[self.context.get_configs()["presets"][preset_name]["provider"]][0]
        else:
            class_name = self.context.get_configs()["presets"][preset_name]["class_name"]
            module_name=self.context.get_configs()["presets"][preset_name]["module_name"]

        try:
            module = __import__(module_name, fromlist=[class_name])
        except KeyError:
            module = __import__(module_name, fromlist=[class_name])
        if hasattr(module, class_name):
            return LlmConcurrencyProxy(LlmLoggingProxy(getattr(module, class_name)(self.context.get_configs()["presets"][preset_name])))
        else:
            raise Exception("Class not found")



from setuptools import setup, find_packages

setup(
    name="ext_llm",
    version="0.1.10",
    description="A wrapper library to abstract common llm providers",
    author="Giovanni Pio Grieco",
    author_email="gio.grieco@stud.uniroma3.it",
    license="GPL3",
    packages=find_packages(),  # Automatically find and include all packages
    install_requires=[
        # List your dependencies here
        "pyyaml",
        "boto3",
        "botocore",
        "groq"
    ],
    url="https://github.com/giovanni-grieco/ext_llm",
)


from concurrent.futures import Future


class Response:
    def __init__(self, content, metadata):
        self.content = content
        self.metadata = metadata

    def __str__(self) -> str:
        return self.content.__str__()


class Stream:
    def __init__(self):
        pass

    def __iter__(self):
        raise NotImplementedError("This method should be implemented by subclasses.")

class Llm:
    def __init__(self):
        pass

    def generate_text(self, system_prompt : str, prompt : str, max_tokens=None, temperature=None) -> Future[Stream | Response] :
        raise NotImplementedError("This method should be implemented by subclasses.")

    def get_config(self):
        raise NotImplementedError("This method should be implemented by subclasses.")

